/*

DESCRIPTION:
    Main source file for 12VDC->5VDC SMPS non-isolated buck based on full-bridge
    inverter development using PIC16F1769.

SYNTAX:  n/a

KNOWN BUGS:  n/a

COPYRIGHT:  (c) 2017, Brian Cornell, all rights reserved.

DEPENDENCIES:
    main.h - configuration constants for main.c

VERSION:  1.1, production release

MODIFICATION HISTORY:

    28 Oct 17, BWC, created file.
    

*/


#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <pic.h>
#include <xc.h>
#include <binary.h>
#include "config.h"
#include "pval.h"
#include "main.h"
#include "array.h"



/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *                              Data objects
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
const uint16_t *sineArray;      //Sine array to use based on bridge volts
uint8_t sineSelect = 0;         //Selects pointer to sine array
unsigned char sineIndex;        //Index to active element in array pointed to by sineArray
bit sineDir,                    //Sine angle increasing/decreasing
        statRun,                //Bridge status - run/stop
        cmdStop;                //Commands ISR to finish current halversine & stop

#ifdef FEATURE_NTC
bit tempHOT, tempCHK;           //Flags to indicate temperature status
#endif



/* Function main.  Initialization & runtime control loop.
 *
 * Returns:  none
 *
 * Parameters:  none
 *
*/
void main(void)
{
#ifdef FEATURE_REG
    union {
        unsigned char result[ADC_SIZ_INT];
        unsigned int value;
    } adc;
unsigned int auxVolts, sampleVolts;
#endif
    uint8_t
#ifdef FEATURE_REG
            sample = S_IN_SAMPLE,
            sineHysteresis = 0,             //Hysteresis for sineSelect
            sineCount = 0,                  //Hysteresis counter for sineSelect
#endif
            timer = 0;

#ifdef FEATURE_NTC
    uint8_t hysHOT, hysOVH;                 //Temp sense hysteresis counters
#endif

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Boot system and pause for settle time.
     +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    Initialize();
    __delay_ms(S_TM_INT);

#ifdef FEATURE_NTC
    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * If temp feature enabled initialize control objects.
     * 
     * Do a one-time check (e.g. without hysteresis) for high temp to set 
     * tempHOT TRUE.  Better to prevent a start right after boot than allow only 
     * to immediately shut down.
     +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    tempHOT = FALSE;
    tempCHK = S_BL_HOT;
    hysHOT = 0;
    hysOVH = 0;

    if (S_MP_HOT) tempHOT = TRUE;
#endif

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     *                     ***** Run-time loop *****
     * Enable watchdog timer.
     +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    SWDTEN = TRUE;
    while (TRUE)
    {
        CLRWDT();                               //Clear watchdog timer

#ifdef FEATURE_REG
        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * Measure auxiliary PS voltage.  For efficiency the raw ADC value is
         * used.
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        ADCON0bits.GO = TRUE;                       //Start conversion
        while (ADCON0bits.GO);                      //Wait for completion
        adc.result[ADC_POS_HIG] = ADRESH;           //Fetch result
        adc.result[ADC_POS_LOW] = ADRESL;
        sampleVolts += adc.value;

        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * Use auxiliary PS voltage to select the appropriate sine array to
         * maintain a semi-regulated 120 VAC RMS output.  To minimize jitter a 
         * number of samples (generally exceeding 1 60Hz cycle, 16.666mS) are 
         * taken, averaged, and scaled.
         * 
         * Bounds checking is performed first and the array address assigned to
         * sineSelect as necessary.  Otherwise the array offset is calculated
         * based on the difference of minimum & actual voltage.
         * 
         * The bounds checked quotient represents the offset in listArray which
         * contains the list of sine arrays and is stored in object sineSelect.
         * Since listArray is constant and sineSelect is a single byte, the start
         * function and ISR select & update (respectively) the sine table in use 
         * without conflict or risk of data corruption (during instant of 
         * exception).
         * 
         * Other actions may be added for low/high limits.
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        if (++sample > S_LM_AUXSAMPLES)
        {
            //Calculate average
            auxVolts = (sampleVolts / S_LM_AUXSAMPLES);

            //Evaluate for selection of appropriate sine array to use
            if (auxVolts < S_VL_AUXMIN) sineHysteresis = 0;
            else if (auxVolts > S_VL_AUXMAX) sineHysteresis = S_LM_SINETABLE;
            else sineHysteresis = ((auxVolts - S_VL_AUXMIN) / S_VL_AUXINC);

            if (sineHysteresis != sineSelect) ++sineCount; else if (sineCount)\
                --sineCount;
            if (sineCount > S_LM_SINE_HYS)
            {
                sineCount = 0;
                sineSelect = sineHysteresis;
            }

            sample = S_IN_SAMPLE;                   //Reset sample counter
            sampleVolts = 0;                        //Clear for new round of samples
        }
#endif


#ifdef FEATURE_NTC
        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * Monitor temp if feature enabled.  Alternate between HOT and OverHeat 
         * checks.
         * 
         * Behavior:
         * 1) If controller boots with temperature above HOT threshold it
         * will not start.  Likewise, if already on and temperature is above HOT
         * threshold prior to CTRL set TRUE it won't start.
         * 2) When running and temperature rises above HOT threshold the ACK
         * pin is set FALSE but unit remains on.
         * 3) When temperature rises above OVH threshold the unit is turned off
         * and will not resume operation, provided CTRL remains TRUE, until 
         * temperature falls below HOT threshold.
         * 
         * The Digital to Analog Converter (DAC) and Comparator are used for
         * temperature monitoring.  The NTC is expected to be the bottom 
         * resistor in the voltage divider so as temperature rises the voltage
         * on the Comparator's (-) pin will fall.
         * 
         * During normal operation the DAC is set to the HOT threshold and the
         * comparator's output monitored.  When TRUE the algorithm monitors
         * both the HOT and OVH thresholds by toggling the DAC's value.
         * 
         * Hysteresis is employed for noise filtering.  Successive samples
         * increment the hysteresis counter but any gap in positive samples 
         * decrements thereby extending the required sample time to confirm (a
         * simple PI loop).
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        if (tempHOT)
        {
            if (tempCHK == S_BL_OVH)        //Check for overheating
            {
                if (S_MP_OVH)
                {
                    if (++hysOVH > S_LM_OVH_HYS)
                    {
                        sineOFF();          //Overheat threshold confirmed:  turn off
                        hysOVH = 0;
                    }
                }
                else if (hysOVH) --hysOVH;

                if (!(hysOVH))              //Toggle to HOT temp check
                {
                    tempCHK = S_BL_HOT;
                    SETDACHOT();
                }
            }
            else                            //Check if unit has cooled down
            {
                if (!(S_MP_HOT))
                {
                    if (++hysHOT > S_LM_HOT_HYS)
                    {
                        tempHOT = FALSE;            //Confirmed, resume operation
                        hysHOT = 0;
                        if (statRun) SETACKON();    //Clear fault
                    }
                }
                else if (hysHOT) --hysHOT;

                //Only check for overheating if SSR is still on
                if (!(hysHOT) && (tempHOT) && (statRun))
                {
                    tempCHK = S_BL_OVH;
                    SETDACOVH();
                }
            }
        }
        else
        {
            if (S_MP_HOT)                   //Hot temp check
            {
                if (++hysHOT > S_LM_HOT_HYS)
                {
                    hysHOT = 0;             //Confirmed, next check for overheat
                    tempHOT = TRUE;
                    tempCHK = S_BL_OVH;
                    SETDACOVH();
                    SETACKOFF();            //If on keep running but signal high temp
                }
            }
            else if (hysHOT) --hysHOT;
        }
#endif


        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * Turn AC output on:  start sine wave generation.
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#ifdef FEATURE_NTC
        if (((PIN_I_RUN == S_CD_RUN) && (statRun == S_CD_STOP)) && !(tempHOT))
#else
        if ((PIN_I_RUN == S_CD_RUN) && (statRun == S_CD_STOP))
#endif
        {
            //Start after hysteresis time to reduce false turn-on
            if (++timer > S_TM_TIMEOUT)
            {
                sineON();
                timer = 0;                      //Clear timer
            }
            else continue;
        }

        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * Turn AC output off:  stop sine wave generation.
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        if ((PIN_I_RUN == S_CD_STOP) && (statRun == S_CD_RUN))
        {
            if (++timer > S_TM_TIMEOUT) sineOFF();
            else continue;
        }

        //If bottom of loop reached then no timing in progress so clear
        timer = 0;
    }
} //End function main


/* Function sineON.  Turns sine output on.
 * 
 * FYI:  in a conventional full bridge that utilizes a bootstrap capacitor 
 * arrangement to support the high-side gate drive there is usually some 
 * accommodation for charging the capacitors prior to starting.  This can be done
 * in the gate driver HW or controller SW.
 * 
 * Because of the unique operating scheme employed by the sine bridge this isn't
 * necessary:  the duty cycle is expected to start at or near zero and hence for
 * most of the 1st/2nd periods the same side low will be on for most of that 
 * time which provides ample charge time.  It could be argued that that during
 * the first cycle there will be some minimal on-time (perhaps a few hundred ns)
 * and that the capacitors won't contain sufficient charge to drive the gate, 
 * resulting in a missed cycle and/or excessive losses.  Due to the short on-time 
 * the impact of this is considered inmaterial and is a one-time event.
 * 
 * Returns:  none
 *
 * Parameters:  none
 *
*/
void sineON()
{
    uint8_t count;

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Select waveform amplitude based on current auxiliary volts, start
     * at 0 degrees with bridge A high-side active.
     +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    sineArray = listArray[sineSelect];
    sineIndex = 0;
    sineDir = S_BL_INCR;
    PIN_O_ENA = TRUE;
    PIN_O_ENB = FALSE;

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Initialize timer & PWM, load starting duty cycle (DC).
     +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    TMR2 = 0;                       //Clear timer count
    TMR2IF = FALSE;                 //Clear timer interrupt flag
    PWM5DCH = (uint8_t) (*sineArray >> CFG_PWM_DCH_BITS);
    PWM5DCL = (uint8_t) ((*sineArray & CFG_PWM_DCL_MASK) << CFG_PWM_DCL_BITS);
    PWM5EN = TRUE;
    TMR2ON = TRUE;

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Let timer2 run for S_LM_TMR2START periods before starting CWG to
     * ensure a clean start (0 angle) for sine wave.
     * 
     * Enable timer interrupt and set acknowledge pin & status.  ISR 
     * continues sine pattern in background.
     +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    for (count = 0; (count < S_LM_TMR2START); )
    {
        if (TMR2IF)
        {
            ++count;
            TMR2IF = FALSE;
        }
    }
    CWG1AS0bits.SHUTDOWN = FALSE;
    TMR2IE = TRUE;
    SETACKON();
    statRun = S_CD_RUN;

    return;
} //End function sineON


/* Function sineOFF.  Turns sine output off.
 * 
 * Issue command & wait for ISR to clear, then clear acknowledge pin & run status.
 * 
 * Returns:  none
 *
 * Parameters:  none
 *
*/
void sineOFF()
{
    cmdStop = TRUE;                 //Issue stop command
    while (cmdStop);                //Wait
    SETACKOFF();                    //Clear acknowledge pin
    statRun = S_CD_STOP;            //Set status

    return;
} //End function sineOFF


/*Function InterruptExec.  MPU interrupt handler.  Manages all enabled
 * interrupts:  timer2.
 *
 * Returns:  none
 *
 * Parameters:  none
 *
*/
void interrupt InterruptExec(void)
{
    static unsigned int dc;

    if (TMR2IF)
    {
        //Fetch next DC
        dc = *(sineArray + sineIndex);
        //Load next duty cycle value
        PWM5DCH = (uint8_t) (dc >> CFG_PWM_DCH_BITS);
        PWM5DCL = (uint8_t) ((dc & CFG_PWM_DCL_MASK) << CFG_PWM_DCL_BITS);

        //Check direction & if at peak reverse
        if (sineDir == S_BL_INCR)
        {
            if (sineIndex < S_LM_PWMINDEX) ++sineIndex;
            else sineDir = S_BL_DECR;
        }
        else
        {
            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             * When index reaches zero reverse polarity and begin next 
             * halversine.  If stop bit set disable CWG, PWM, timer, and clear 
             * interrupt enable to prevent further generation.
             * 
             * Stop is acknowledged by clearing cmdStop bit prior to return.
             +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            if (sineIndex > 0) --sineIndex;
            else
            {
                if (cmdStop)
                {
                    CWG1AS0bits.SHUTDOWN = TRUE;    //Stop CWG
                    PIN_O_ENA = FALSE;              //Clear bridge enable pins
                    PIN_O_ENB = FALSE;
                    PWM5EN = FALSE;                 //Stop PWM
                    TMR2ON = FALSE;                 //Stop timer
                    TMR2IE = FALSE;                 //Prevent timer interrupts
                    cmdStop = FALSE;                //Acknowledge stop
                }
                else
                {
                    //Reverse polarity every 180 degrees
                    sineDir = S_BL_INCR;

#ifdef FEATURE_REG
                    //Always update the sine table (just as fast as checking 1st)
                    sineArray = listArray[sineSelect];
#endif

                    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
                     * The high side output must be off (e.g. duty cycle 
                     * completed) before direction bits can be flipped to 
                     * prevent shoot-thru, so Execution is suspended until the 
                     * PWM output is FALSE.  Since the DC at 0 or 180 degrees is 
                     * nearly zero the wait time will be very small (<1uS).
                     +++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
                    while (S_BL_HSON);
                    PIN_O_ENA ^= TRUE;
                    PIN_O_ENB ^= TRUE;
                }
            }
        }

        TMR2IF = FALSE;
    }

    return;
} //End function InterruptExec


/*Function Initialize.  Performs baseline MPU initialization at boot.
 *
 * Returns:  none
 *
 * Parameters:  none
 *
*/
void Initialize(void)
{
    di();                               //No interrupts during initialization

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Configure system clock & watchdog.  Expects 1MHz HFINTOSC to be active at 
     * boot.  Adjust HF internal oscillator to obtain 60Hz period from TMR2.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    OSCCON1bits.NOSC = CFG_OSC_SRC;
    OSCCON1bits.NDIV = CFG_OSC_DIV;
    OSCFRQ = CFG_OSC_FRQ;
#ifndef _SIMUL
    while (!(OSCSTAT1bits.HFOR));       //Make sure oscillator is online
#endif
    OSCTUNE = CFG_OSC_OSCTUNE;          //Shift oscillator frequency if necessary
    WDTCON = CFG_WDT_CON;               //Set watchdog timer interval


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * If booting for anything other than power-up halt for safety.  This will
     * prevent erroneous / unsafe operation of unit and force operator 
     * intervention.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    if (((PCON0 & S_LM_PCON_EX) != S_LM_PCON_BT) || \
            ((STATUS & S_LM_STAT_EX) != S_LM_STAT_BT)) while (TRUE);


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Important but miscellaneous settings that must be set at boot and aren't
     * part of a larger group.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    sineDir = S_BL_INCR;
    statRun = S_CD_STOP;
    cmdStop = FALSE;


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Set port configuration.  To prevent any aberrant behaviors before 
     * initialization is complete set output latches.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    //Port A
    PORTA = VL_PORTA;
    TRISA = IO_PORTA;
    ANSELA = AD_PORTA;
    WPUA = WP_PORTA;
    ODCONA = OD_PORTA;
    SLRCONA = SR_PORTA;
    INLVLA = IL_PORTA;
    //Port C
    PORTC = VL_PORTC;
    TRISC = IO_PORTC;
    ANSELC = AD_PORTC;
    WPUC = WP_PORTC;
    ODCONC = OD_PORTC;
    SLRCONC = SR_PORTC;
    INLVLC = IL_PORTC;


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Map peripheral pins
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    //CWG & CLC outputs
    PPS_P_CWG1_A = PPS_M_CWG1_A;
    PPS_P_CWG1_B = PPS_M_CWG1_B;
    PPS_P_CLC1 = PPS_M_CLC1;
    PPS_P_CLC2 = PPS_M_CLC2;
    PPS_P_CLC3 = PPS_M_CLC3;
    PPS_P_CLC4 = PPS_M_CLC4;

    //CLC inputs
    CLCIN0PPS = PPS_M_CLC_0;
    CLCIN1PPS = PPS_M_CLC_1;
    CLCIN2PPS = PPS_M_CLC_2;
    CLCIN3PPS = PPS_M_CLC_3;


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Configure FVR & DAC
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    FVRCON = CFG_FVR_C;
    DACCON0 = CFG_DAC_CON0;
    DACCON1 = CFG_DAC_HOT;


#ifdef FEATURE_REG
    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Configure ADC
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    ADCON1 = CFG_ADC_CON1;
    ADCON0 = CFG_ADC_CON0;
#endif


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Configure CM1 (comparator)
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    CM1CON0 = CFG_CM_CON0;
    CM1CON1 = CFG_CM_CON1;


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Configure Timer2
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    T2CON = CFG_TMR_2_CON;
    PR2 = CFG_TMR_2_FRQ;
    
    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Configure PWM5
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    PWM5CON = CFG_PWM_5_CON;
    PWMTMRS = CFG_PWM_5_TMRS;

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Configure CWG1.  During config CWG should be disabled.  Once done auto-
     * shutdown bit should be set to prevent spurious operation so unit can be 
     * enabled.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    CWG1CON0 = CFG_CWG_1_CON0;
    CWG1CON1 = CFG_CWG_1_CON1;
    CWG1CLKCON = CFG_CWG_1_CLK;
    CWG1DAT = CFG_CWG_1_DAT;
    CWG1AS0 = CFG_CWG_1_AS0;
    CWG1AS1 = CFG_CWG_1_AS1;
    CWG1DBR = CFG_CWG_1_DBR;
    CWG1DBF = CFG_CWG_1_DBF;
    CWG1CON0bits.EN = TRUE;


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Configure CLCs.  Once configured enable.  This ensures output remains low 
     * (dependent on CWG being configured and placed in AS state).
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    //CLC1
    CLC1CON = CFG_CLC_1_CON;
    CLC1POL = CFG_CLC_1_POL;
    CLC1SEL0 = CFG_CLC_1_SEL0;
    CLC1SEL1 = CFG_CLC_1_SEL1;
    CLC1SEL2 = CFG_CLC_1_SEL2;
    CLC1SEL3 = CFG_CLC_1_SEL3;
    CLC1GLS0 = CFG_CLC_1_LS_0;
    CLC1GLS1 = CFG_CLC_1_LS_1;
    CLC1GLS2 = CFG_CLC_1_LS_2;
    CLC1GLS3 = CFG_CLC_1_LS_3;
    CLC1CONbits.EN = TRUE;

    //CLC2
    CLC2CON = CFG_CLC_2_CON;
    CLC2POL = CFG_CLC_2_POL;
    CLC2SEL0 = CFG_CLC_2_SEL0;
    CLC2SEL1 = CFG_CLC_2_SEL1;
    CLC2SEL2 = CFG_CLC_2_SEL2;
    CLC2SEL3 = CFG_CLC_2_SEL3;
    CLC2GLS0 = CFG_CLC_2_LS_0;
    CLC2GLS1 = CFG_CLC_2_LS_1;
    CLC2GLS2 = CFG_CLC_2_LS_2;
    CLC2GLS3 = CFG_CLC_2_LS_3;
    CLC2CONbits.EN = TRUE;

    //CLC3
    CLC3CON = CFG_CLC_3_CON;
    CLC3POL = CFG_CLC_3_POL;
    CLC3SEL0 = CFG_CLC_3_SEL0;
    CLC3SEL1 = CFG_CLC_3_SEL1;
    CLC3SEL2 = CFG_CLC_3_SEL2;
    CLC3SEL3 = CFG_CLC_3_SEL3;
    CLC3GLS0 = CFG_CLC_3_LS_0;
    CLC3GLS1 = CFG_CLC_3_LS_1;
    CLC3GLS2 = CFG_CLC_3_LS_2;
    CLC3GLS3 = CFG_CLC_3_LS_3;
    CLC3CONbits.EN = TRUE;

    //CLC4
    CLC4CON = CFG_CLC_4_CON;
    CLC4POL = CFG_CLC_4_POL;
    CLC4SEL0 = CFG_CLC_4_SEL0;
    CLC4SEL1 = CFG_CLC_4_SEL1;
    CLC4SEL2 = CFG_CLC_4_SEL2;
    CLC4SEL3 = CFG_CLC_4_SEL3;
    CLC4GLS0 = CFG_CLC_4_LS_0;
    CLC4GLS1 = CFG_CLC_4_LS_1;
    CLC4GLS2 = CFG_CLC_4_LS_2;
    CLC4GLS3 = CFG_CLC_4_LS_3;
    CLC4CONbits.EN = TRUE;


    PEIE = ENABLE;                      //Enable peripheral interrupts
    ei();                               //Enable interrupts

    return;
} //End function Initialize



//End of file